"""
Import of  default state and extras
"""
from mr.states.implementations.memory import MemoryState

__all__ = ["MemoryState"]

# Redis edition extra
try:
    import redis.asyncio as redis
    from .implementations.redis import RedisState

    __all__ += "RedisState"
except ImportError as error:  # pragma: no cover
    pass

# Temp edition extra
try:
    from aiofile import async_open
    from .implementations.temp import TempFileState

    __all__ += "TempFileState"
except ImportError as error:  # pragma: no cover
    pass



"""
State interface. You can implement your own State approach
"""

from abc import ABC, abstractmethod


class IState(ABC):
    """
    State interface
    """

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def sync_get(self, key: int):
        """
        Sync get implementation
        :param key: int
        :return:
        """

    @abstractmethod
    def sync_set(self, key: int, value: any, ttl: int):
        """
        Sync set implementation
        :param key: int
        :param value: Any
        :param ttl: int. Seconds that the cache will have to live. Set None to never die
        :return:
        """

    @abstractmethod
    async def async_get(self, key: int):
        """
        Async get implementation
        :param key: int
        :return:
        """

    @abstractmethod
    async def async_set(self, key: int, value: any, ttl: int):
        """
        Async set implementation
        :param key: int
        :param value: Any
        :param ttl: int. Seconds that the cache will have to live. Set None to never die
        :return:
        """



"""
Memory state implementation
"""
import pickle
import tempfile
from datetime import datetime
from inspect import _empty
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

import aiofile

from mr.states.interface import IState


class TempFileState(IState):
    """
    State interface
    """

    _path: Path
    _temp_folder: TemporaryDirectory

    def __init__(self, **kwargs):
        try:
            _dir = Path(tempfile.gettempdir())
            if kwargs.get("BASE_PATH") is not None:
                _dir = Path(kwargs["BASE_PATH"])
            if kwargs.get("STATIC"):
                self._path = _dir
            else:
                self._temp_folder = TemporaryDirectory(  # pylint: disable=R1732
                    dir=_dir
                )
                self._path = Path(self._temp_folder.name)
        except (TypeError, FileNotFoundError, PermissionError) as exception:
            raise TypeError(
                "The config value BASE_PATH is not a valid Path"
            ) from exception

    def get_path(self, key: int) -> Path:
        """
        Get temp file path
        :return: Path
        """
        temp_path = self._path / str(key)
        return temp_path

    def sync_get(self, key: int):
        path = self.get_path(key=key)
        if path.exists():
            with open(path, "rb") as file:
                value = file.read()
            return self.__unpickle_with_header(value)
        return None

    def sync_set(self, key: int, value: any, ttl: int = _empty):
        path = self.get_path(key=key)
        with open(path, "wb") as file:
            file.write(self.__pickle_with_header(value, ttl))

    async def async_get(self, key: int):
        path = self.get_path(key=key)
        if path.exists():
            async with aiofile.async_open(path, "rb") as file:
                value = await file.read()
            return self.__unpickle_with_header(value)
        return None

    async def async_set(self, key: int, value: any, ttl: int = _empty):
        path = self.get_path(key=key)
        async with aiofile.async_open(path, "wb") as file:
            await file.write(self.__pickle_with_header(value, ttl))

    @staticmethod
    def __pickle_with_header(value: any, ttl: int = _empty) -> bytes:
        pack = {
            "created_at": datetime.utcnow().timestamp(),
            "ttl": ttl,
            "value": pickle.dumps(value),
        }
        return pickle.dumps(pack)

    @staticmethod
    def __unpickle_with_header(pack_bytes: bytes) -> Optional[any]:
        now_timestamp = datetime.utcnow().timestamp()
        pack = pickle.loads(pack_bytes)
        register_timestamp = pack.get("created_at")
        _ttl = pack.get("ttl")
        if _ttl == _empty or ((register_timestamp + _ttl) >= now_timestamp):
            return pickle.loads(pack.get("value"))
        return None



"""
Mr Mime is a function decorator for cache/memoization
"""
import functools
import inspect
import itertools
from inspect import _empty
from typing import TypeVar, Hashable

import meeseeks

from mr.config import Config
from mr.states.implementations.memory import MemoryState

T = TypeVar("T")

singleton_container = meeseeks.OnlyOne(by_args_hash=True)


@singleton_container
class Mime:
    """
    Decorator to aplay cache/memoization on your functions
    """

    _config: Config = Config(state=MemoryState)

    @classmethod
    def set_config(cls, config: Config):
        """
        Replace the default config

        :param config: Config. IS a instance os Config class
        :return: None
        """
        cls._config = config

    def __init__(self, ttl: int = _empty):
        self._ttl = ttl

    @staticmethod
    def _hash_args(func_hash: int, args: tuple, kwargs: dict) -> int:
        """
        Created for each arg + kwargs hash. The kwargs`s order doesn't have influence
        """
        hash_instance = tuple(
            item if isinstance(item, Hashable) else str(item)
            for item in itertools.chain(
                [func_hash], args, [i[1] for i in sorted(kwargs.items())]
            )
        )
        return hash(hash_instance)

    def __call__(self, callable_obj: T) -> T:
        _is_class = inspect.isclass(callable_obj)
        if inspect.isclass(callable_obj):
            raise TypeError(
                "Mr. Mime don`t support class memoization for this use meeseeks-singleton package\n"
                "link: https://pypi.org/project/meeseeks-singleton/"
            )
        _func_hash = hash(callable_obj)
        if inspect.iscoroutinefunction(callable_obj):

            @functools.wraps(callable_obj)
            async def async_mimic(*args, **kwargs):
                args_hash = self._hash_args(
                    func_hash=_func_hash, args=args, kwargs=kwargs
                )
                async with self._config.async_acquire_state() as state:
                    if cached_value := await state.async_get(key=args_hash):
                        return cached_value
                    value = await callable_obj(*args, **kwargs)
                    await state.async_set(key=args_hash, value=value, ttl=self._ttl)
                    return value

            return async_mimic

        @functools.wraps(callable_obj)
        def sync_mimic(*args, **kwargs):
            args_hash = self._hash_args(func_hash=_func_hash, args=args, kwargs=kwargs)
            with self._config.sync_acquire_state() as state:
                if cached_value := state.sync_get(key=args_hash):
                    return cached_value
                value = callable_obj(*args, **kwargs)
                state.sync_set(key=args_hash, value=value, ttl=self._ttl)
                return value

        return sync_mimic



"""
Memory state implementation
"""
from datetime import datetime
from inspect import _empty

from mr.states.interface import IState


class MemoryState(IState):
    """
    State that use hash table to save cached returns
    """

    _state: dict[int, any]
    _kwargs: dict[str, any]

    __slots__ = ("_state", "_kwargs")

    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self._state = {}

    def sync_get(self, key: int):
        now_timestamp = datetime.utcnow().timestamp()
        if register := self._state.get(key):
            _ttl = register.get("ttl")
            if _ttl == _empty or ((register.get("created_at") + _ttl) >= now_timestamp):
                return register.get("value")
        return None

    def sync_set(self, key: int, value: any, ttl: int = _empty):
        value = {
            "created_at": datetime.utcnow().timestamp(),
            "ttl": ttl,
            "value": value,
        }
        self._state.update({key: value})

    async def async_get(self, key: int):
        return self.sync_get(key)

    async def async_set(self, key: int, value: any, ttl: int = _empty):
        return self.sync_set(key=key, value=value, ttl=ttl)



"""
Config for Mr Mime
"""
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from typing import Type

from mr.states.interface import IState


@dataclass(slots=True, frozen=True)
class Config:
    """
    Config class
    """

    state: Type[IState]
    state_kwargs: dict = field(default_factory=dict)
    _state: IState = None

    def __post_init__(self):
        if not issubclass(self.state, IState):
            raise TypeError("State class does not implement IState interface")

    @property
    def initialized_state(self):
        """
        Initialized sate
        @return:
        """
        if self._state is None:
            object.__setattr__(self, "_state", self.state(**self.state_kwargs))
        return self._state

    @contextmanager
    def sync_acquire_state(self) -> IState:
        """
        Get sync state
        :return: SyncRedis
        """
        yield self.initialized_state

    @asynccontextmanager
    async def async_acquire_state(self) -> IState:
        """
        Get sync state
        :return: SyncRedis
        """
        yield self.initialized_state
